--
-- ForestArea
--
-- # Author:  Heady
-- # date: 24.02.2019
-- # last update: 08.05.2020
--
-- > Not for commercial use  < --
-- > Not allowed for edit - Copyright (C) Heady - www.planet-ls.de < --
--


local modDir = g_currentModDirectory;

local modDescXML = loadXMLFile("modDesc", modDir .. "modDesc.xml");
local version = getXMLString(modDescXML, "modDesc.version");

print("Script: ForestArea v"..tostring(version).." loaded by Heady");


ModEvent = {};

ModEvent.forestAreaObjects = {};

function ModEvent:loadMap(name)

	self.tickRate = 30;
	self.tickDuration = self.tickRate/1000;
	self.tickSum = 0;
	self.updateTime = 0;	
	self.allowUpdateTickRoutine = false;
	
end;

function ModEvent:deleteMap()

	ModEvent.forestAreaObjects = {};

end;

function ModEvent:delete()
end;

function ModEvent:mouseEvent(posX, posY, isDown, isUp, button)
end;

function ModEvent:keyEvent(unicode, sym, modifier, isDown)
end;

function ModEvent:update(dt)

	self.tickSum = self.tickSum + dt;
	if self.tickDuration - 3 <= self.tickSum then
		self.allowUpdateTickRoutine = true;
		for index, object in pairs(ModEvent.forestAreaObjects) do
			if object ~= nil then
				if object.updateTick ~= nil and object.finalizePlaced ~= nil then
					if object.finalizePlaced then
						object:updateTick(self.tickSum);
					end;
				end;
			end;
		end;		
		self.tickSum = 0;
	end;
	self.allowUpdateTickRoutine = false;
	
end;

function ModEvent:draw()

end;

addModEventListener(ModEvent);



function onCreate(self, id)

end;

ForestArea = {};

ForestArea.treesPerHaPerDay = 13;

local ForestArea_mt = Class(ForestArea, Object);
InitObjectClass(ForestArea, "ForestArea");

function ForestArea:new(isServer, isClient, customMt)
    local mt = customMt;
    if mt == nil then
        mt = ForestArea_mt;
    end;
	local instance = Object:new(isServer, isClient, mt);
	registerObjectClassName(instance, "ForestArea");

	return instance;
end;

function ForestArea:load(id, xmlFile)

	self.nodeId = id;
	
	
	self.forestAreas = {};
	if hasXMLProperty(xmlFile, "placeable.forestAreas") then
        local i = 0;
        while true do
            local key = string.format("placeable.forestAreas.forestArea(%d)", i);
            if not hasXMLProperty(xmlFile, key) then
                break;
            end;
			local entry = {};
			entry.startNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, key .. "#startNode"));
			entry.widthNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, key .. "#widthNode"));
			entry.heightNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, key .. "#heightNode"));
			table.insert(self.forestAreas, entry);
            i = i + 1;
        end;
	end;
	
	--self.treesPerHaPerDay = getXMLFloat(xmlFile, "placeable.forestAreas#treesPerHaPerDay");

	self.runOnce = true;
	
	self.plantTime = 0;
	self.lastTreePos = {};
	self.mapTreesPos = {};
	self.forestAreaSize = 0;
	
	self.treeCounter = 0;

	
	self.mpDirtyFlag = self:getNextDirtyFlag();	
	self.serverUpdate = false;
	
	table.insert(ModEvent.forestAreaObjects, self);
	return true;
end;

function ForestArea:delete()

	
	for index, object in pairs(ModEvent.forestAreaObjects) do
		if object ~= nil then
			if object == self then
				table.remove(ModEvent.forestAreaObjects, index);
			end;
		end;
	end;

end;

function ForestArea:loadFromXMLFile(xmlFile, key, resetVehicles)

	
    return true;
end;

function ForestArea:saveToXMLFile(xmlFile, key, usedModNames)
	
end;

function ForestArea:mouseEvent(posX, posY, isDown, isUp, button)
end;

function ForestArea:keyEvent(unicode, sym, modifier, isDown)
end;

--[[function ForestArea:readStream(streamId, connection)
    if connection:getIsServer() then

    end
end;

function ForestArea:writeStream(streamId, connection)
    if not connection:getIsServer() then

    end;
end;

function ForestArea:readUpdateStream(streamId, timestamp, connection)
	if connection:getIsServer() then

	end;
end;

function ForestArea:writeUpdateStream(streamId, connection, dirtyMask)
	if not connection:getIsServer() then
	
	end;
end;]]


function ForestArea:update(dt)

	if self.runOnce then
		self.forestAreaSize = self:calcAreaSizeHa(self.forestAreas);
		
		self.runOnce = false;
	end;

end;

function ForestArea:updateTick(dt)

	if g_server ~= nil then
	
		local treePlantRate = 1/( (ForestArea.treesPerHaPerDay*self.forestAreaSize)/(24*60) ); -- minutes per tree per ha

		self.plantTime = self.plantTime + ((g_currentMission.missionInfo.timeScale*dt)/(1000*60));
		
		if self.plantTime > treePlantRate * 3 then
			self.plantTime = 0;
		end;

		if self.plantTime > treePlantRate then
			if g_treePlantManager:canPlantTree() then
				local randomI = math.random(table.getn(self.forestAreas));			
				--for i=1, randomI do
					local area = self.forestAreas[randomI];
					local xs,ys,zs = getWorldTranslation(area.startNode);
					local xw,yw,zw = getWorldTranslation(area.widthNode);
					local xh,yh,zh = getWorldTranslation(area.heightNode);

					local ux, uy, uz = xw-xs, yw-ys, zw-zs; -- vektor width
					local vx, vy, vz = xh-xs, yh-ys, zh-zs; -- vektor height

					local px = xs + (math.random()*ux) + (math.random()*vx); -- random point X
					local py = ys + (math.random()*uy) + (math.random()*vy); -- random point Y
					local pz = zs + (math.random()*uz) + (math.random()*vz); -- random point Z

					local height = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, px,py,pz);					
					local yRot = math.random() * 2*math.pi;		
					
					--TreePlantUtil.plantTree(g_currentMission.plantedTrees, TreePlantUtil.treeTypes["treeFir"].index, px,height,pz, 0,yRot,0, 0);
					local canPlantTree = self:canCreateTree(px, height, pz);
					if canPlantTree then
						g_treePlantManager:plantTree(1, px, height, pz, 0, yRot, 0, 0);
						self.treeCounter = self.treeCounter + 1;
						--print("tree planted #"..tostring(self.treeCounter).."  "..px.."|"..height.."|"..pz)
						self.lastTreePos = {px,height,pz};
						self.plantTime = 0;
					end;
				--end;
			end;
		end;
		
	end;


	if self.serverUpdate then
		self:raiseDirtyFlags(self.mpDirtyFlag);
		self.serverUpdate = false;
	end;		

end;

function ForestArea:draw()

end;

function ForestArea:setOwnerFarmId(farmId)

	if farmId == nil then
        farmId = AccessHandler.EVERYONE;
    end;
	self.ownerFarmId = farmId;

end;

function ForestArea:getOwnerFarmId()
	return self.ownerFarmId;
end;


local bitMapSize = 16384;

function ForestArea:calcAreaSizeHa(areas)

	local sumPixel = 0;
	local bitVector = createBitVectorMap("forestArea");
	loadBitVectorMapNew(bitVector, bitMapSize, bitMapSize, 1, true);
	for i=1, table.getn(areas) do
		--print("calcAreaSize i: "..i)
		local x,_,z = getWorldTranslation(areas[i].startNode);
		local widthX,_,widthZ = getWorldTranslation(areas[i].widthNode);
		local heightX,_,heightZ = getWorldTranslation(areas[i].heightNode);

		local x,z = self:convertWorldToFieldOwnershipPosition(x,z);
		local widthX,widthZ = self:convertWorldToFieldOwnershipPosition(widthX,widthZ);
		local heightX,heightZ = self:convertWorldToFieldOwnershipPosition(heightX,heightZ);

		sumPixel = sumPixel + setBitVectorMapParallelogram(bitVector, x, z, widthX-x, widthZ-z, heightX-x, heightZ-z, 0, 1, 0);
		--print("sumPixel i: "..sumPixel)
	end;
	local areaSizeHa = self:pixelToHa(sumPixel);
	--print(string.format("%s : %.3f ha", "forestArea", areaSizeHa))


	delete(bitVector);
	return areaSizeHa;
end;

function ForestArea:convertWorldToFieldOwnershipPosition(x,z)
	--print("g_currentMission.terrainRootNode: "..tostring(g_currentMission.terrainRootNode))
	local terrainSize = getTerrainSize(g_currentMission.terrainRootNode);
	--print("convert terrainSize: "..terrainSize)
    return math.floor(bitMapSize * (x+terrainSize*0.5) / terrainSize), math.floor(bitMapSize * (z+terrainSize*0.5) / terrainSize);
end

function ForestArea:pixelToHa(area)
	--print("g_currentMission.terrainRootNode: "..tostring(g_currentMission.terrainRootNode))
	local terrainSize = getTerrainSize(g_currentMission.terrainRootNode);
	--print("pixel terrainSize: "..terrainSize)
    local pixelToSqm = terrainSize/bitMapSize;
    return (area*pixelToSqm*pixelToSqm) / 10000;
end

function ForestArea:canCreateTree(x, y, z, farmId)
	local canPlantTree = true;
    
    if not g_treePlantManager:canPlantTree() then
        --self.showTooManyTreesWarning = true;
        canPlantTree = false;
		--print("g_treePlantManager:canPlantTree()")
		return false;
    end
	
	self.findObjectsCount = 0;
	local collisionMask = 4294967040;
	local numShapesSphere = overlapSphere(x, y, z, 1.5, "findObjectsCallback", self, collisionMask, true, true, false)
	if self.findObjectsCount > 0 then
		canPlantTree = false;
		return false;
	end;
	
	if g_treePlantManager.treesData ~= nil and g_treePlantManager.treesData.growingTrees ~= nil then
		for key, value in pairs(g_treePlantManager.treesData.growingTrees) do
			if value.x ~= nil and value.y ~= nil and value.z ~= nil then

				local distance = MathUtil.vector2LengthSq(x - value.x, z - value.z)
				if distance < 1.5 then
					--print("distance < 1.5")
					canPlantTree = false;
					return false;
				end;
			end
		end
	end;
	if g_treePlantManager.treesData ~= nil and g_treePlantManager.treesData.splitTrees ~= nil then
		for key, value in pairs(g_treePlantManager.treesData.splitTrees) do
			if value.x ~= nil and value.y ~= nil and value.z ~= nil then

				local distance = MathUtil.vector2LengthSq(x - value.x, z - value.z)
				if distance < 1.5 then
					--print("distance < 1.5")
					canPlantTree = false;
					return false;
				end;
			end
		end
	end;
		
	return true;	
end;

function ForestArea:findObjectsCallback(transformId)
    if transformId ~= 0 and getHasClassId(transformId, ClassIds.SHAPE) then
		self.findObjectsCount = self.findObjectsCount + 1;
		--print("findObjectsCallback(), getCollisionMask(transformId): "..tostring(getCollisionMask(transformId)).." getName(transformId): "..getName(transformId).." getName(getParent(transformId)): "..getName(getParent(transformId)))
    end
    return true
end

